from contextlib import suppress


class SourceGroup:
    def __init__(self, /, **kwargs):
        self.groups = kwargs
        if assignments := kwargs.pop('assign_outputs', None):
            for grp, output in assignments.items():
                for signal in self.groups[grp]:
                    signal.preferred_out = output

    def get_companions(self, source):
        idx = None
        for sourceset in self.groups.values():
            with suppress(ValueError):
                idx = sourceset.index(source)
                break

        if idx is not None:
            rv = set()
            for sourceset in self.groups.values():
                with suppress(IndexError):
                    companion = sourceset[idx]
                    if companion is not source:
                        rv.add(companion)
            return rv

        return set()


class MatrixGroup:
    def __init__(self, signalgrp, **kwargs):
        self.signals = signalgrp
        self.matricies = kwargs

    def select(self, matrix, idx, src, no_companions=False):
        mat = self.matricies[matrix]
        mat.select(idx, src)
        mat_out = mat.outputs[idx]

        if not no_companions:
            companions = self.signals.get_companions(src)
            for other_src in companions:
                outp = other_src.preferred_out
                if outp and outp is not mat_out:
                    outp.select(other_src, nolock=True)
                else:
                    print(f"skipping {other_src}, no pref output")

    def get_output(self, name, idx: int):
        return self.matricies[name].outputs[idx]



# signals,py - Information and different video streams in the code
from typing import Protocol
import uuid


class Source:
    def __init__(self, name, preferred_out=None):
        self.uuid = uuid.uuid4()
        self.name = name
        self.preferred_out = preferred_out

    def __repr__(self):
        return f"Source({self.name})"


class Sink(Protocol):
    def source_changed(source) -> None:
        pass



from unittest import TestCase, skip
from unittest.mock import Mock
from worchestic.matrix import (
    Matrix,
    MatrixOutput,
    LockedOutput,
    AlreadyUnlocked,
)

from utils import make_signal


class MatrixOuputTests(TestCase):
    def setUp(self):
        self.idx = 0
        self.o = MatrixOutput(Mock(), self.idx)

    def test_output_select_forwards_call_to_matrix(self):
        source = make_signal()
        self.o.select(source)
        self.o._device._select.assert_called_with(self.idx, source)

    def test_output_select_locks_output(self):
        source = make_signal()
        self.o.select(source)
        self.assertTrue(self.o.locked)

    def test_output_select_sets_sources_atttribute(self):
        source = make_signal()
        self.o.select(source)
        self.assertEqual(self.o._source, source)

    def test_output_select_doesnt_lock_output_with_nolock(self):
        source = make_signal()
        self.o.select(source, nolock=True)
        self.assertFalse(self.o.locked)

    def test_claiming_an_output_locks_it(self):
        self.o.claim()
        self.assertTrue(self.o.locked)

    def test_selecting_on_a_locked_outout_raises_LockedOutput(self):
        self.o.select(make_signal())
        with self.assertRaises(LockedOutput):
            self.o.select(make_signal())

    def test_releasing_an_output_allows_reassignment(self):
        self.o.claim()
        self.o.release()
        self.o.select(make_signal())

    def test_unlock_doesnt_happen_if_release_arent_balanced(self):
        self.o.claim()
        self.o.claim()
        self.o.release()
        self.assertTrue(self.o.locked)

    def test_unlock_happens_if_release_are_balanced(self):
        self.o.claim()
        self.o.claim()
        self.o.release()
        self.o.release()
        self.assertFalse(self.o.locked)

    def test_releasing_an_unlocked_output_raises_AlreadyUnlocked(self):
        with self.assertRaises(AlreadyUnlocked):
            self.o.release()


class SimpleMatrixTests(TestCase):
    def setUp(self):
        self.driver = Mock()
        self.sources = [make_signal(), make_signal()]
        self.m = Matrix("simple", self.driver, self.sources, 2)

    def test_matrix_is_truthy(self):
        self.assertTrue(self.m)

    def test_aavailable_source_with_no_cascade(self):
        self.assertSetEqual(self.m.available_sources, set(self.sources))

    def test_select_on_single_matrix_calls_the_driver(self):
        self.m.select(1, self.sources[0])
        self.driver.select.assert_called_with(1, 0)

    def test_can_construct_matrix_cascades(self):
        sources2 = [make_signal(), make_signal()]
        m2 = Matrix("m2", Mock(), sources2, 2)
        master = Matrix("master", Mock(), self.m.outputs + m2.outputs, 2)
        self.assertTrue(master)

    def test_select_on_single_matrix_updates_the_source_of_the_output(self):
        self.m.select(1, self.sources[0])
        self.assertEqual(self.m.outputs[1].source, self.sources[0])


class TwoLevelMatrixTests(TestCase):
    """Two level matrix checks the basic interactions
    bwtween matrix an outputs as inputs"""
    def setUp(self):
        self.sources1 = [make_signal(), make_signal(), make_signal()]
        self.m1 = Matrix("m1", Mock(), self.sources1, 2)
        # Share the middle source on both leaf matricies
        self.sources2 = [self.sources1[1], make_signal()]
        self.m2 = Matrix("m2", Mock(), self.sources2, 2)
        self.root_m = Matrix("root",
                             Mock(),
                             self.m1.outputs + self.m2.outputs, 3)

    def test_available_source_removes_repeated_options(self):
        self.assertSetEqual(self.root_m.available_sources,
                            set(self.sources1 + [self.sources2[1]]))

    def test_iter_sources_has_pathlen2_and_all_sources(self):
        sources = list(self.root_m.iter_sources())
        self.assertEqual(len(sources), 10)
        self.assertEqual([s.path_len for s in sources], [2] * 10)

    def test_selecting_a_source_at_the_root_selects_it_at_the_leaf(self):
        self.root_m.select(0, self.sources1[1])
        # Actaully  the driver being called with either (0,0), or (0,1)
        # is valid but it depends on the routing algo which one
        # is actually chosen - the current code chooses (0,0)
        self.root_m._driver.select.assert_called_with(0, 0)
        self.m1._driver.select.assert_called_with(0, 1)

    def test_selecting_a_second_source_doesnt_reuse_a_path(self):
        self.root_m.select(0, self.sources1[1])
        self.root_m._driver.select.assert_called_with(0, 0)
        self.m1._driver.select.assert_called_with(0, 1)
        self.root_m.select(1, self.sources1[0])
        self.root_m._driver.select.assert_called_with(1, 1)
        self.m1._driver.select.assert_called_with(1, 0)

    def tie_up_m1(self):
        # Tie up all the outputs from _m1.
        self.root_m.select(0, self.sources1[1])
        self.root_m._driver.select.assert_called_with(0, 0)
        self.m1._driver.select.assert_called_with(0, 1)
        self.root_m.select(1, self.sources1[0])
        self.root_m._driver.select.assert_called_with(1, 1)
        self.m1._driver.select.assert_called_with(1, 0)

    def test_selecting_an_unroutable_source_raises_an_exception(self):
        self.tie_up_m1()
        # Select a source only available on an m1 output
        with self.assertRaises(Exception):
            self.root_m.select(2, self.sources1[2])

    def test_reselecting_an_ouptut_works_if_resource_are_swapped(self):
        self.tie_up_m1()
        # Select a source only available on an m1 output
        self.root_m.select(1, self.sources1[2])

    def test_reselecting_an_ouptut_releases_uinused_resources(self):
        self.tie_up_m1()
        # Select a source only available on an m1 output
        self.root_m.select(1, self.sources2[1])
        self.assertEqual(
                1,
                # Count locked outputs
                sum(1 for o in self.m1.outputs if o.locked)
        )

    def test_selecting_an_ouptut_as_a_mirror_claims_and_existing_input(self):
        self.tie_up_m1()
        # Select an alreay in use source
        self.root_m.select(0, self.sources1[0])
        self.assertFalse(self.m1.outputs[0].locked)
        self.assertEqual(self.m1.outputs[1]._sem.load(), 2)


class ThreeLevelMatrixTests(TestCase):
    """A three level system is need to test
    recursion happens correctly""
    """
    def setUp(self):
        self.sources1 = [make_signal(), make_signal(), make_signal()]
        self.m1 = Matrix("m1", Mock(), self.sources1, 2)
        # Share the middle source on both leaf matricies
        self.sources2 = [self.sources1[1], make_signal()]
        self.m2 = Matrix("m2", Mock(), self.sources2, 2)
        self.sources3 = [make_signal(), make_signal()]
        self.m3 = Matrix("m3", Mock(), self.sources3, 2)
        self.n1 = Matrix("n1",
                         Mock(),
                         self.m1.outputs + [self.m2.outputs[0]], 2)

        self.n2 = Matrix("n2",
                         Mock(),
                         self.m3.outputs + [self.m2.outputs[1]], 2)

        self.root_m = Matrix("root",
                             Mock(),
                             self.n1.outputs + self.n2.outputs, 3)

    def test_selecting_in_input_recursively_claims_outputs(self):
        self.root_m.select(0, self.sources1[0])
        self.assertEqual(
                1,
                # Count locked outputs
                sum(1 for o in self.m1.outputs if o.locked)
        )
        self.assertEqual(
                1,
                # Count locked outputs
                sum(1 for o in self.n1.outputs if o.locked)
        )

    def test_a_selecting_a_different_input_recursively_releases_outputs(self):
        self.root_m.select(0, self.sources1[0])
        self.root_m.select(0, self.sources3[0])
        self.assertEqual(
                0,
                # Count locked outputs
                sum(1 for o in self.n1.outputs if o.locked)
        )
        self.assertEqual(
                0,
                # Count locked outputs
                sum(1 for o in self.m1.outputs if o.locked)
        )

    def test_updating_a_selected_input_source_cascades_to_the_output(self,):
        self.root_m.select(0, self.sources1[0])
        self.assertEqual(self.root_m.outputs[0].uuid, self.sources1[0].uuid)
        self.m1.replug_input(0, self.sources3[-1])
        self.assertEqual(self.root_m.outputs[0].uuid, self.sources3[-1].uuid)



from unittest import TestCase, skip
from unittest.mock import Mock, patch, call
from utils import make_signal
from worchestic.matrix import Matrix
from worchestic.group import MatrixGroup, SourceGroup


class SourceGroupTest(TestCase):
    def setUp(self):
        self.video = [make_signal(), make_signal(), make_signal()]
        self.usb = [make_signal(), make_signal(), make_signal()]
        self.signal_group = SourceGroup(
                    video_a=self.video[:2],
                    video_b=self.video[2:],
                    usb=self.usb
        )

    def test_get_companions_returns_the_empty_list_for_unknown_sources(self):
        self.assertEqual(self.signal_group.get_companions(make_signal), set())

    def test_given_a_source_its_companions_can_be_retrieved(self):
        companions_0 = set([self.video[2], self.usb[0]])
        self.assertSetEqual(set(self.signal_group.get_companions(self.video[0])),
                            companions_0)


class SingalGroupInitTests(TestCase):
    def test_assign_out_set_preferred_ouptut(self):
        self.video = [make_signal(), make_signal(), make_signal(), make_signal()]
        self.usb = [make_signal(), make_signal(), make_signal(), make_signal()]
        self.mat_video = Matrix("video", Mock(), self.video, 2)
        self.mat_usb = Matrix("usb", Mock(), self.usb, 1)
        self.signal_group = SourceGroup(
                    video_a=self.video[:2],
                    video_b=self.video[2:],
                    usb=self.usb,
                    assign_outputs={
                        'usb': self.mat_usb.outputs[0],
                        'video_a': self.mat_video.outputs[0],
                        'video_b': self.mat_video.outputs[1],

                    }
        )
        for i, signal in enumerate(self.video):
            self.assertEqual(
                signal.preferred_out, self.mat_video.outputs[i // 2]
            )
        for i, signal in enumerate(self.usb):
            self.assertEqual(signal.preferred_out, self.mat_usb.outputs[0])



class MatrixGroupTest(TestCase):
    def setUp(self):
        self.video = [make_signal(), make_signal(), make_signal(), make_signal()]
        self.usb = [make_signal(), make_signal(), make_signal(), make_signal()]
        self.signal_group = SourceGroup(
                    video_a=self.video[:2],
                    video_b=self.video[2:],
                    usb=self.usb
        )
        self.mat_video = Matrix("video", Mock(), self.video, 2)
        self.mat_usb = Matrix("usb", Mock(), self.usb, 1)
        for hid in self.usb:
            hid.preferred_out = self.mat_usb.outputs[0]

        for idx, screen in enumerate(self.video):
            screen.preferred_out = self.mat_video.outputs[idx // 2]

        self.mgroup = MatrixGroup(
                self.signal_group,
                usb=self.mat_usb, video=self.mat_video
        )

    def test_selecting_a_source_selects_it_in_the_matrix_and_nothing_else_if_no_companions_is_true(self):
        with patch.object(self.mat_video, "_select") as vid_sel:
            self.mgroup.select("video", 0, self.video[0], no_companions=True)

        self.mat_usb._driver.assert_not_called()
        vid_sel.assert_called_once_with(0, self.video[0])

    def test_selecting_a_source_selects_it_in_the_matrix_and_companions_is_default(self):
        with patch.object(self.mat_video, "_select") as vid_sel, \
             patch.object(self.mat_usb, "_select") as usb_sel:
            self.mgroup.select("video", 0, self.video[0])

        usb_sel.assert_called_once_with(0, self.usb[0])
        vid_sel.assert_has_calls([
                call(0, self.video[0]),
                call(1, self.video[2]),
        ])

    def test_selecting_a_source_selects_it_in_the_matrix_and_skip_companons_for_the_chosen_output(self):
        with patch.object(self.mat_video, "_select") as vid_sel, \
             patch.object(self.mat_usb, "_select") as usb_sel:
            self.mgroup.select("video", 1, self.video[0])

        usb_sel.assert_called_once_with(0, self.usb[0])
        vid_sel.assert_called_once_with(1, self.video[0])

    def test_get_outpots_returns_indexed_ouptut_from_named_matrix(self):
        mname = "video"
        index = 1
        self.assertIs(
            self.mgroup.get_output(mname, index),
            self.mat_video.outputs[index]
        )



# vid_matrix.py - Video matrix controll logic

from contextlib import suppress
from dataclasses import dataclass
from typing import Union, ForwardRef
from .signals import Source, Sink
from atomicx import AtomicInt
import logging

logger = logging.getLogger(__name__)


class LockedOutput(ValueError):
    pass


class AlreadyUnlocked(ValueError):
    pass


class UnroutableOutput(ValueError):
    pass


class MatrixOutput:
    """A Matrix output is a source, which is
    the output of a specrific discoverable matrix"""
    def __init__(self, device: ForwardRef('Matrix'),
                 idx: int, sink: Sink = None):
        self._source = None
        self._sem = AtomicInt(0)
        self._device = device
        self._idx = idx
        self.connection = None

    def connected_to(self, sink):
        self.connection = sink

    def _source_changed(self, source):
        if self._source is not source:
            self._source = source
            self.connection and self.connection.source_changed(source)

    def __str__(self):
        return f"{self._device.name}.outputs[{self._idx}]"

    @property
    def port(self):
        return self._device, self._idx

    @property
    def locked(self):
        return bool(self._sem.load())

    @property
    def source(self):
        return self._source

    @property
    def uuid(self):
        """Unique Identifier for the current signal on this output/cable"""
        return self._source and self._source.uuid

    def select(self, src: ForwardRef('InputSignal'), nolock: bool = False):
        """Select an alternate signal for this ouptut, locking the output"""
        if src.uuid != self.uuid:
            if self.locked:
                raise LockedOutput(f"{self} is locked/in use")
            self._device._select(self._idx, src)
            self._source_changed(src)
        if not nolock:
            self.claim()

    def release(self):
        """Release a lock on the output"""
        self._sem.dec()
        if self._sem.load() < 0:
            # Attempt to recover!
            self._sem.inc()
            raise AlreadyUnlocked("Invalid lock state")
        self._device.release(self._idx)

    def claim(self):
        """Claim a lock on the output"""
        self._sem.inc()


class MatrixDriver:
    def select(input: int, output: int):
        raise NotImplementedError()


InputSignal = Union[MatrixOutput, Source]


class Matrix:
    """An instance of this class
    represents a single switch element"""
    class Input:
        def __init__(self, matrix, idx):
            self.matrix = matrix
            self.idx = idx

        def source_changed(self, src):
            self.matrix._input_changed(self.idx, src)

    def __init__(self, name: str, driver: MatrixDriver, inputs: InputSignal,
                 nr_outputs: int):
        self.name = name
        self._driver = driver
        self.inputs = inputs
        self.outputs = [None] * nr_outputs
        self._current = {}
        for idx in range(nr_outputs):
            self.outputs[idx] = MatrixOutput(self, idx)

        for idx, source in enumerate(self.inputs):
            if isinstance(source, MatrixOutput):
                source.connected_to(self.Input(self, idx))

    def __str__(self):
        return self.name

    @dataclass
    class AvailableSource:
        input_idx: int
        path_len: int
        path: InputSignal
        source: Source

    @property
    def available_sources(self):
        """List the available sources for this matrix"""
        return set(x.source for x in self.iter_sources())

    def iter_sources(self):
        for idx, inp in enumerate(self.inputs):
            if isinstance(inp, MatrixOutput):
                if inp.locked:
                    # If the input is locked don't recurse.
                    # but it is availibl itself.
                    logger.debug(f"locked - {idx}, {inp}")
                    yield self.AvailableSource(idx,  0, inp, inp._source)
                    continue

                for source in inp.port[0].iter_sources():
                    logger.debug(f"unlocked - {idx}, {source.source.uuid}")
                    yield self.AvailableSource(idx, source.path_len + 1,
                                               inp, source.source)
            else:
                yield self.AvailableSource(idx, 1, inp, inp)

    def _input_changed(self, idx, source):
        for out, inp in self._current.items():
            #if self.outputs[out].locked:
            #    raise LockedOutput(f"Input {idx} is used by locked output {self.outputs[out]}")
            if idx == inp:
                self.outputs[out]._source_changed(source)

    def replug_input(self, idx, source):
        """Changes the input found on a source"""
        self._input_changed(idx, source)
        self.inputs[idx] = source

    def select(self, idx, source: Source):
        """Sets output (idx) to connect to source

        Propagates up the switch fabric as necessary.
        """
        self.outputs[idx].select(source, nolock=True)

    def _select(self, idx, source: Source):
        "internal select function"
        logger.info(f"{self}: assigning {idx} to {source.uuid}")
        self.release(idx)

        routes = [s for s in self.iter_sources() if s.source == source]
        if not routes:
            raise UnroutableOutput(f"{self}:{source.uuid} is not routable to output {idx}")

        route = min(routes, key=lambda s: s.path_len)
        if isinstance(route.path, MatrixOutput):
            logger.info(f"({self})Using output {route.path}({route.path_len}) "
                        f"for {idx}")
            route.path.select(route.source)
        self._driver.select(idx, route.input_idx)
        self._current[idx] = route.input_idx

    def release(self, idx):
        """Releases a hold on any signal apatch which feed this input

           - Allows the currently used input for this output to be switched
             to a differnet signal
        """
        try:
            current = self.inputs[self._current[idx]]
            current.release()
            logger.debug(f"released {current and current.uuid}")
        except (KeyError, AttributeError) as e:
            logger.debug(f"skipping release: {e!r}")
